@@ -32,8 +32,8 @@ function splitter(array, filter) {
 
 // Functions that start with '$' should not be exported to the wasm module.
 // They are intended to be exclusive to JS code only.
-function isJsOnlyIdentifier(ident) {
-  return ident[0] == '$';
+function isJsOnlySymbol(symbol) {
+  return symbol[0] == '$';
 }
 
 function escapeJSONKey(x) {
@@ -66,41 +66,49 @@ function isDefined(symName) {
   return false;
 }
 
-// JSifier
-function runJSify(functionsOnly) {
-  const mainPass = !functionsOnly;
-  const functionStubs = [];
+function getTransitiveDeps(symbol) {
+  // TODO(sbc): Use some kind of cache to avoid quadratic behaviour here.
+  const transitiveDeps = new Set();
+  const seen = new Set();
+  const toVisit = [symbol];
+  while (toVisit.length) {
+    const sym = toVisit.pop();
+    if (!seen.has(sym)) {
+      let directDeps = LibraryManager.library[sym + '__deps'] || [];
+      directDeps = directDeps.filter((d) => typeof d === 'string');
+      if (directDeps.length) {
+        directDeps.forEach(transitiveDeps.add, transitiveDeps);
+        toVisit.push(...directDeps);
+      }
+      seen.add(sym);
+    }
+  }
+  return Array.from(transitiveDeps);
+}
 
-  const itemsDict = {type: [], functionStub: [], function: [], globalVariablePostSet: []};
+function runJSify() {
+  const libraryItems = [];
+  const symbolDeps = {};
+  const asyncFuncs = [];
+  let postSets = [];
 
-  if (mainPass) {
-    // Add additional necessary items for the main pass. We can now do this since types are parsed (types can be used through
-    // generateStructInfo in library.js)
+  LibraryManager.load();
 
-    LibraryManager.load();
-
-    const libFuncsToInclude = DEFAULT_LIBRARY_FUNCS_TO_INCLUDE;
-    for (const sym of EXPORTED_RUNTIME_METHODS) {
-      if ('$' + sym in LibraryManager.library) {
-        libFuncsToInclude.push('$' + sym);
+  const symbolsNeeded = DEFAULT_LIBRARY_FUNCS_TO_INCLUDE;
+  for (const sym of EXPORTED_RUNTIME_METHODS) {
+    if ('$' + sym in LibraryManager.library) {
+      symbolsNeeded.push('$' + sym);
+    }
+  }
+  if (INCLUDE_FULL_LIBRARY) {
+    for (const key of Object.keys(LibraryManager.library)) {
+      if (!isJsLibraryConfigIdentifier(key)) {
+        symbolsNeeded.push(key);
       }
     }
-    if (INCLUDE_FULL_LIBRARY) {
-      for (const key in LibraryManager.library) {
-        if (!isJsLibraryConfigIdentifier(key)) {
-          libFuncsToInclude.push(key);
-        }
-      }
-    }
-    libFuncsToInclude.forEach((ident) => {
-      functionStubs.push({
-        identOrig: ident,
-        identMangled: mangleCSymbolName(ident),
-      });
-    });
   }
 
-  function convertPointerParams(snippet, sig) {
+  function convertPointerParams(symbol, snippet, sig) {
     // Automatically convert any incoming pointer arguments from BigInt
     // to double (this limits the range to int53).
     // And convert the return value if the function returns a pointer.
@@ -112,6 +120,10 @@ function runJSify(functionsOnly) {
       let argConvertions = '';
       for (let i = 1; i < sig.length; i++) {
         const name = argNames[i - 1];
+        if (!name) {
+          error(`convertPointerParams: missing name for argument ${i} in ${symbol}`);
+          return snippet;
+        }
         if (sig[i] == 'p') {
           argConvertions += `  ${name} = Number(${name});\n`;
           newArgs.push(`Number(${name})`);
@@ -142,7 +154,7 @@ ${argConvertions}
     });
   }
 
-  function processLibraryFunction(snippet, ident, finalName, deps) {
+  function processLibraryFunction(snippet, symbol, mangled, deps, isStub) {
     // It is possible that when printing the function as a string on Windows,
     // the js interpreter we are in returns the string with Windows line endings
     // \r\n. This is undesirable, since line endings are managed in the form \n
@@ -151,13 +163,17 @@ ${argConvertions}
     snippet = snippet.toString().replace(/\r\n/gm, '\n');
 
     // name the function; overwrite if it's already named
-    snippet = snippet.replace(/function(?:\s+([^(]+))?\s*\(/, 'function ' + finalName + '(');
+    snippet = snippet.replace(/function(?:\s+([^(]+))?\s*\(/, 'function ' + mangled + '(');
+
+    if (isStub) {
+      return snippet;
+    }
 
     // apply LIBRARY_DEBUG if relevant
-    if (LIBRARY_DEBUG && !isJsOnlyIdentifier(ident)) {
+    if (LIBRARY_DEBUG && !isJsOnlySymbol(symbol)) {
       snippet = modifyFunction(snippet, (name, args, body) => `\
 function ${name}(${args}) {
-  var ret = (function() { if (runtimeDebug) err("[library call:${finalName}: " + Array.prototype.slice.call(arguments).map(prettyPrint) + "]");
+  var ret = (function() { if (runtimeDebug) err("[library call:${mangled}: " + Array.prototype.slice.call(arguments).map(prettyPrint) + "]");
   ${body}
   }).apply(this, arguments);
   if (runtimeDebug && typeof ret != "undefined") err("  [     return:" + prettyPrint(ret));
@@ -165,10 +181,37 @@ function ${name}(${args}) {
 }`);
     }
 
+    if (SHARED_MEMORY) {
+      const proxyingMode = LibraryManager.library[symbol + '__proxy'];
+      if (proxyingMode) {
+        if (proxyingMode !== 'sync' && proxyingMode !== 'async') {
+          throw new Error(`Invalid proxyingMode ${symbol}__proxy: '${proxyingMode}' specified!`);
+        }
+        const sync = proxyingMode === 'sync';
+        if (PTHREADS) {
+          snippet = modifyFunction(snippet, (name, args, body) => `
+function ${name}(${args}) {
+if (ENVIRONMENT_IS_PTHREAD)
+  return proxyToMainThread(${proxiedFunctionTable.length}, ${+sync}${args ? ', ' : ''}${args});
+${body}
+}\n`);
+        } else if (WASM_WORKERS && ASSERTIONS) {
+          // In ASSERTIONS builds add runtime checks that proxied functions are not attempted to be called in Wasm Workers
+          // (since there is no automatic proxying architecture available)
+          snippet = modifyFunction(snippet, (name, args, body) => `
+function ${name}(${args}) {
+  assert(!ENVIRONMENT_IS_WASM_WORKER, "Attempted to call proxied function '${name}' in a Wasm Worker, but in Wasm Worker enabled builds, proxied function architecture is not available!");
+  ${body}
+}\n`);
+        }
+        proxiedFunctionTable.push(mangled);
+      }
+    }
+
     if (MEMORY64) {
-      const sig = LibraryManager.library[ident + '__sig'];
+      const sig = LibraryManager.library[symbol + '__sig'];
       if (sig && sig.includes('p')) {
-        snippet = convertPointerParams(snippet, sig);
+        snippet = convertPointerParams(symbol, snippet, sig);
       }
     }
 
@@ -178,13 +221,16 @@ function ${name}(${args}) {
   function addImplicitDeps(snippet, deps) {
     // There are some common dependencies that we inject automatically by
     // conservatively scanning the input functions for their usage.
-    // Specifically, these are dependencies that are automatically generated by
-    // the {{{ makeDynCall }}}, and {{{ runtimeKeepalivePush/Pop }}} macros.
+    // Specifically, these are dependencies that are very common and would be
+    // burdensome to add manually to all functions.
+    // The first four are deps that are automatically/conditionally added
+    // by the {{{ makeDynCall }}}, and {{{ runtimeKeepalivePush/Pop }}} macros.
     const autoDeps = [
       'getDynCaller',
       'getWasmTableEntry',
       'runtimeKeepalivePush',
       'runtimeKeepalivePop',
+      'UTF8ToString',
     ];
     for (const dep of autoDeps) {
       if (snippet.includes(dep + '(')) {
@@ -193,127 +239,167 @@ function ${name}(${args}) {
     }
   }
 
-  // functionStub
-  function functionStubHandler(item) {
+  function symbolHandler(symbol) {
     // In LLVM, exceptions generate a set of functions of form
     // __cxa_find_matching_catch_1(), __cxa_find_matching_catch_2(), etc.  where
     // the number specifies the number of arguments. In Emscripten, route all
     // these to a single function '__cxa_find_matching_catch' that variadically
     // processes all of these functions using JS 'arguments' object.
-    if (item.identMangled.startsWith('___cxa_find_matching_catch_')) {
+    if (!WASM_EXCEPTIONS && symbol.startsWith('__cxa_find_matching_catch_')) {
       if (DISABLE_EXCEPTION_THROWING) {
         error('DISABLE_EXCEPTION_THROWING was set (likely due to -fno-exceptions), which means no C++ exception throwing support code is linked in, but exception catching code appears. Either do not set DISABLE_EXCEPTION_THROWING (if you do want exception throwing) or compile all source files with -fno-except (so that no exceptions support code is required); also make sure DISABLE_EXCEPTION_CATCHING is set to the right value - if you want exceptions, it should be off, and vice versa.');
         return;
       }
-      const num = +item.identMangled.split('_').slice(-1)[0];
+      const num = +symbol.split('_').slice(-1)[0];
       addCxaCatch(num);
       // Continue, with the code below emitting the proper JavaScript based on
       // what we just added to the library.
     }
 
-    const TOP_LEVEL = 'top-level compiled C/C++ code';
-
-    function addFromLibrary(item, dependent) {
+    function addFromLibrary(symbol, dependent, force = false) {
       // dependencies can be JS functions, which we just run
-      if (typeof item == 'function') return item();
-
-      const ident = item.identOrig;
-      const finalName = item.identMangled;
-
-      if (ident in addedLibraryItems) return '';
-      addedLibraryItems[ident] = true;
-
-      // don't process any special identifiers. These are looked up when processing the base name of the identifier.
-      if (isJsLibraryConfigIdentifier(ident)) {
-        return '';
+      if (typeof symbol == 'function') {
+        return symbol();
       }
 
-      // if the function was implemented in compiled code, there is no need to include the js version
-      if (WASM_EXPORTS.has(ident)) {
-        return '';
+      // don't process any special identifiers. These are looked up when
+      // processing the base name of the identifier.
+      if (isJsLibraryConfigIdentifier(symbol)) {
+        return;
       }
 
+      const deps = LibraryManager.library[symbol + '__deps'] || [];
+      let sig = LibraryManager.library[symbol + '__sig'];
+      if (!WASM_BIGINT && sig && sig[0] == 'j') {
+        // Without WASM_BIGINT functions that return i64 depend on setTempRet0
+        // to return the upper 32-bits of the result.
+        // See makeReturn64 in parseTools.py.
+        deps.push('setTempRet0');
+      }
+
+      let isAsyncFunction = false;
+      if (ASYNCIFY) {
+        const original = LibraryManager.library[symbol];
+        if (typeof original == 'function' ) {
+          isAsyncFunction = LibraryManager.library[symbol + '__async'] ||
+                            original.constructor.name == 'AsyncFunction'
+        }
+        if (isAsyncFunction) {
+          asyncFuncs.push(symbol);
+        }
+      }
+
+      if (symbolsOnly) {
+        if (LibraryManager.library.hasOwnProperty(symbol)) {
+          var value = LibraryManager.library[symbol];
+          var resolvedSymbol = symbol;
+          // Resolve aliases before looking up deps
+          if (typeof value == 'string' && value[0] != '=' && LibraryManager.library.hasOwnProperty(value)) {
+            resolvedSymbol = value;
+          }
+          var transtiveDeps = getTransitiveDeps(resolvedSymbol);
+          symbolDeps[symbol] = transtiveDeps.filter((d) => !isJsOnlySymbol(d) && !(d in LibraryManager.library));
+        }
+        return;
+      }
+
+      // if the function was implemented in compiled code, there is no need to
+      // include the js version
+      if (WASM_EXPORTS.has(symbol) && !force) {
+        return;
+      }
+
+      if (symbol in addedLibraryItems) {
+        return;
+      }
+      addedLibraryItems[symbol] = true;
+
       // This gets set to true in the case of dynamic linking for symbols that
       // are undefined in the main module.  In this case we create a stub that
       // will resolve the correct symbol at runtime, or assert if its missing.
       let isStub = false;
 
-      if (!LibraryManager.library.hasOwnProperty(ident)) {
-        if (ONLY_CALC_JS_SYMBOLS) {
-          return;
-        }
-        const isWeakImport = WEAK_IMPORTS.has(ident);
-        if (!isDefined(ident) && !isWeakImport) {
-          if (PROXY_TO_PTHREAD && !MAIN_MODULE && ident == '__main_argc_argv') {
+      const mangled = mangleCSymbolName(symbol);
+
+      if (!LibraryManager.library.hasOwnProperty(symbol)) {
+        const isWeakImport = WEAK_IMPORTS.has(symbol);
+        if (!isDefined(symbol) && !isWeakImport) {
+          if (PROXY_TO_PTHREAD && !MAIN_MODULE && symbol == '__main_argc_argv') {
             error('PROXY_TO_PTHREAD proxies main() for you, but no main exists');
             return;
           }
-          let undefinedSym = ident;
-          if (ident === '__main_argc_argv') {
+          let undefinedSym = symbol;
+          if (symbol === '__main_argc_argv') {
             undefinedSym = 'main/__main_argc_argv';
           }
           let msg = 'undefined symbol: ' + undefinedSym;
           if (dependent) msg += ` (referenced by ${dependent})`;
           if (ERROR_ON_UNDEFINED_SYMBOLS) {
             error(msg);
-            if (dependent == TOP_LEVEL && !LLD_REPORT_UNDEFINED) {
-              warnOnce('Link with `-sLLD_REPORT_UNDEFINED` to get more information on undefined symbols');
-            }
             warnOnce('To disable errors for undefined symbols use `-sERROR_ON_UNDEFINED_SYMBOLS=0`');
-            warnOnce(finalName + ' may need to be added to EXPORTED_FUNCTIONS if it arrives from a system library');
+            warnOnce(mangled + ' may need to be added to EXPORTED_FUNCTIONS if it arrives from a system library');
           } else if (VERBOSE || WARN_ON_UNDEFINED_SYMBOLS) {
             warn(msg);
           }
-          if (ident === '__main_argc_argv' && STANDALONE_WASM) {
+          if (symbol === '__main_argc_argv' && STANDALONE_WASM) {
             warn('To build in STANDALONE_WASM mode without a main(), use emcc --no-entry');
           }
         }
         if (!RELOCATABLE) {
           // emit a stub that will fail at runtime
-          LibraryManager.library[ident] = new Function(`err('missing function: ${ident}'); abort(-1);`);
+          LibraryManager.library[symbol] = new Function(`err('missing function: ${symbol}'); abort(-1);`);
           // We have already warned/errored about this function, so for the purposes of Closure use, mute all type checks
           // regarding this function, marking ot a variadic function that can take in anything and return anything.
           // (not useful to warn/error multiple times)
-          LibraryManager.library[ident + '__docs'] = '/** @type {function(...*):?} */';
+          LibraryManager.library[symbol + '__docs'] = '/** @type {function(...*):?} */';
         } else {
-          const target = `Module['${finalName}']`;
+          // Create a stub for this symbol which can later be replaced by the
+          // dynamic linker.  If this stub is called before the symbol is
+          // resolved assert in debug builds or trap in release builds.
+          if (ASYNCIFY) {
+            // See the definition of asyncifyStubs in preamble.js for why this
+            // is needed.
+            target = `asyncifyStubs['${symbol}']`;
+          } else {
+            target = `wasmImports['${symbol}']`;
+          }
           let assertion = '';
           if (ASSERTIONS) {
-            assertion += `if (!${target}) abort("external symbol '${ident}' is missing. perhaps a side module was not linked in? if this function was expected to arrive from a system library, try to build the MAIN_MODULE with EMCC_FORCE_STDLIBS=1 in the environment");\n`;
+            assertion += `if (!${target} || ${target}.stub) abort("external symbol '${symbol}' is missing. perhaps a side module was not linked in? if this function was expected to arrive from a system library, try to build the MAIN_MODULE with EMCC_FORCE_STDLIBS=1 in the environment");\n`;
           }
           const functionBody = assertion + `return ${target}.apply(null, arguments);`;
-          LibraryManager.library[ident] = new Function(functionBody);
+          LibraryManager.library[symbol] = new Function(functionBody);
           isStub = true;
         }
       }
 
-      const original = LibraryManager.library[ident];
+      librarySymbols.push(mangled);
+
+      const original = LibraryManager.library[symbol];
       let snippet = original;
-      const deps = LibraryManager.library[ident + '__deps'] || [];
-      if (!Array.isArray(deps)) {
-        error(`JS library directive ${ident}__deps=${deps.toString()} is of type ${typeof deps}, but it should be an array!`);
-        return;
-      }
-      const isUserSymbol = LibraryManager.library[ident + '__user'];
+
+      const isUserSymbol = LibraryManager.library[symbol + '__user'];
       deps.forEach((dep) => {
         if (typeof snippet == 'string' && !(dep in LibraryManager.library)) {
           warn(`missing library dependency ${dep}, make sure you are compiling with the right options (see #if in src/library*.js)`);
         }
         if (isUserSymbol && LibraryManager.library[dep + '__internal']) {
-          warn(`user library symbol '${ident}' depends on internal symbol '${dep}'`);
+          warn(`user library symbol '${symbol}' depends on internal symbol '${dep}'`);
         }
       });
+
       let isFunction = false;
+      let aliasTarget;
 
       if (typeof snippet == 'string') {
         if (snippet[0] != '=') {
-          const target = LibraryManager.library[snippet];
-          if (target) {
-            // Redirection for aliases. We include the parent, and at runtime make ourselves equal to it.
-            // This avoid having duplicate functions with identical content.
-            const redirectedTarget = snippet;
-            deps.push(redirectedTarget);
-            snippet = mangleCSymbolName(redirectedTarget);
+          if (LibraryManager.library[snippet]) {
+            // Redirection for aliases. We include the parent, and at runtime
+            // make ourselves equal to it.  This avoid having duplicate
+            // functions with identical content.
+            aliasTarget = snippet;
+            snippet = mangleCSymbolName(aliasTarget);
+            deps.push(aliasTarget);
           }
         }
       } else if (typeof snippet == 'object') {
@@ -321,17 +407,11 @@ function ${name}(${args}) {
         addImplicitDeps(snippet, deps);
       } else if (typeof snippet == 'function') {
         isFunction = true;
-        snippet = processLibraryFunction(snippet, ident, finalName, deps);
+        snippet = processLibraryFunction(snippet, symbol, mangled, deps, isStub);
         addImplicitDeps(snippet, deps);
       }
 
-      librarySymbols.push(finalName);
-
-      if (ONLY_CALC_JS_SYMBOLS) {
-        return '';
-      }
-
-      const postsetId = ident + '__postset';
+      const postsetId = symbol + '__postset';
       let postset = LibraryManager.library[postsetId];
       if (postset) {
         // A postset is either code to run right now, or some text we should emit.
@@ -341,55 +421,22 @@ function ${name}(${args}) {
         }
         if (postset && !addedLibraryItems[postsetId]) {
           addedLibraryItems[postsetId] = true;
-          itemsDict.globalVariablePostSet.push({
-            JS: postset + ';',
-          });
+          postSets.push(postset + ';');
         }
       }
 
       if (VERBOSE) {
-        printErr(`adding ${finalName} and deps ${deps} : ` + (snippet + '').substr(0, 40));
+        printErr(`adding ${mangled} and deps ${deps} : ` + (snippet + '').substr(0, 40));
       }
-      const identDependents = ident + "__deps: ['" + deps.join("','") + "']";
+      const deps_list = deps.join("','");
+      const identDependents = symbol + `__deps: ['${deps_list}']`;
       function addDependency(dep) {
-        if (typeof dep != 'function') {
-          dep = {identOrig: dep, identMangled: mangleCSymbolName(dep)};
-        }
-        return addFromLibrary(dep, `${identDependents}, referenced by ${dependent}`);
+        return addFromLibrary(dep, `${identDependents}, referenced by ${dependent}`, dep === aliasTarget);
       }
       let contentText;
       if (isFunction) {
         // Emit the body of a JS library function.
-        const proxyingMode = LibraryManager.library[ident + '__proxy'];
-        if (SHARED_MEMORY && proxyingMode) {
-          if (proxyingMode !== 'sync' && proxyingMode !== 'async') {
-            throw new Error(`Invalid proxyingMode ${ident}__proxy: '${proxyingMode}' specified!`);
-          }
-          const sync = proxyingMode === 'sync';
-          assert(typeof original == 'function');
-          if (USE_PTHREADS) {
-            contentText = modifyFunction(snippet, (name, args, body) => `
-function ${name}(${args}) {
-  if (ENVIRONMENT_IS_PTHREAD)
-    return _emscripten_proxy_to_main_thread_js(${proxiedFunctionTable.length}, ${+sync}${args ? ', ' : ''}${args});
-  ${body}
-}\n`);
-          } else if (WASM_WORKERS) {
-            if (ASSERTIONS) {
-              // In ASSERTIONS builds add runtime checks that proxied functions are not attempted to be called in Wasm Workers
-              // (since there is no automatic proxying architecture available)
-              contentText = modifyFunction(snippet, (name, args, body) => `
-  function ${name}(${args}) {
-    assert(!ENVIRONMENT_IS_WASM_WORKER, "Attempted to call proxied function '${name}' in a Wasm Worker, but in Wasm Worker enabled builds, proxied function architecture is not available!");
-    ${body}
-  }\n`);
-            } else {
-              // In non-ASSERTIONS builds directly emit the original function.
-              contentText = snippet;
-            }
-          }
-          proxiedFunctionTable.push(finalName);
-        } else if ((USE_ASAN || USE_LSAN || UBSAN_RUNTIME) && LibraryManager.library[ident + '__noleakcheck']) {
+        if ((USE_ASAN || USE_LSAN || UBSAN_RUNTIME) && LibraryManager.library[symbol + '__noleakcheck']) {
           contentText = modifyFunction(snippet, (name, args, body) => `
 function ${name}(${args}) {
   return withBuiltinMalloc(function() {
@@ -405,7 +452,7 @@ function ${name}(${args}) {
         //   foo: ';[code here verbatim]'
         //  emits
         //   'var foo;[code here verbatim];'
-        contentText = 'var ' + finalName + snippet;
+        contentText = 'var ' + mangled + snippet;
         if (snippet[snippet.length - 1] != ';' && snippet[snippet.length - 1] != '}') contentText += ';';
       } else {
         // In JS libraries
@@ -415,13 +462,12 @@ function ${name}(${args}) {
         if (typeof snippet == 'string' && snippet[0] == '=') {
           snippet = snippet.substr(1);
         }
-        contentText = `var ${finalName} = ${snippet};`;
+        contentText = `var ${mangled} = ${snippet};`;
       }
-      const sig = LibraryManager.library[ident + '__sig'];
       // asm module exports are done in emscripten.py, after the asm module is ready. Here
       // we also export library methods as necessary.
-      if ((EXPORT_ALL || EXPORTED_FUNCTIONS.has(finalName)) && !isStub) {
-        contentText += `\nModule["${finalName}"] = ${finalName};`;
+      if ((EXPORT_ALL || EXPORTED_FUNCTIONS.has(mangled)) && !isStub) {
+        contentText += `\nModule["${mangled}"] = ${mangled};`;
       }
       // Relocatable code needs signatures to create proper wrappers. Stack
       // switching needs signatures so we can create a proper
@@ -429,36 +475,58 @@ function ${name}(${args}) {
       // TODO: For asyncify we could only add the signatures we actually need,
       //       of async imports/exports.
       if (sig && (RELOCATABLE || ASYNCIFY == 2)) {
-        contentText += `\n${finalName}.sig = '${sig}';`;
+        if (!WASM_BIGINT) {
+          sig = sig[0].replace('j', 'i') + sig.slice(1).replace(/j/g, 'ii');
+        }
+        contentText += `\n${mangled}.sig = '${sig}';`;
+      }
+      if (ASYNCIFY && isAsyncFunction) {
+        contentText += `\n${mangled}.isAsync = true;`;
       }
       if (isStub) {
-        contentText += `\n${finalName}.stub = true;`;
+        contentText += `\n${mangled}.stub = true;`;
+        if (ASYNCIFY) {
+          contentText += `\nasyncifyStubs['${symbol}'] = undefined;`;
+        }
       }
 
       let commentText = '';
-      if (LibraryManager.library[ident + '__docs']) {
-        commentText = LibraryManager.library[ident + '__docs'] + '\n';
+      if (force) {
+        commentText += '/** @suppress {duplicate } */\n'
+      }
+      if (LibraryManager.library[symbol + '__docs']) {
+        commentText += LibraryManager.library[symbol + '__docs'] + '\n';
       }
 
       const depsText = (deps ? deps.map(addDependency).filter((x) => x != '').join('\n') + '\n' : '');
       return depsText + commentText + contentText;
     }
 
-    itemsDict.functionStub.push(item);
-    item.JS = addFromLibrary(item, TOP_LEVEL);
+    const JS = addFromLibrary(symbol, 'top-level compiled C/C++ code');
+    libraryItems.push(JS);
   }
 
-  // Final combiner
+  function includeFile(fileName) {
+    print(`// include: ${fileName}`);
+    print(processMacros(preprocess(fileName)));
+    print(`// end include: ${fileName}`);
+  }
+
+  function includeFileRaw(fileName) {
+    print(`// include: ${fileName}`);
+    print(read(fileName));
+    print(`// end include: ${fileName}`);
+  }
 
   function finalCombiner() {
-    const splitPostSets = splitter(itemsDict.globalVariablePostSet, (x) => x.ident && x.dependencies);
-    itemsDict.globalVariablePostSet = splitPostSets.leftIn;
+    const splitPostSets = splitter(postSets, (x) => x.symbol && x.dependencies);
+    postSets = splitPostSets.leftIn;
     const orderedPostSets = splitPostSets.splitOut;
 
     let limit = orderedPostSets.length * orderedPostSets.length;
     for (let i = 0; i < orderedPostSets.length; i++) {
       for (let j = i + 1; j < orderedPostSets.length; j++) {
-        if (orderedPostSets[j].ident in orderedPostSets[i].dependencies) {
+        if (orderedPostSets[j].symbol in orderedPostSets[i].dependencies) {
           const temp = orderedPostSets[i];
           orderedPostSets[i] = orderedPostSets[j];
           orderedPostSets[j] = temp;
@@ -470,49 +538,30 @@ function ${name}(${args}) {
       }
     }
 
-    itemsDict.globalVariablePostSet = itemsDict.globalVariablePostSet.concat(orderedPostSets);
-
-    //
-
-    if (!mainPass) {
-      const generated = itemsDict.function.concat(itemsDict.type);
-      print(generated.map((item) => item.JS).join('\n'));
-      return;
-    }
+    postSets = postSets.concat(orderedPostSets);
 
     const shellFile = MINIMAL_RUNTIME ? 'shell_minimal.js' : 'shell.js';
+    includeFile(shellFile);
 
-    const shellParts = read(shellFile).split('{{BODY}}');
-    print(processMacros(preprocess(shellParts[0], shellFile)));
-    let pre;
-    if (MINIMAL_RUNTIME) {
-      pre = processMacros(preprocess(read('preamble_minimal.js'), 'preamble_minimal.js'));
-    } else {
-      pre = processMacros(preprocess(read('support.js'), 'support.js')) +
-            processMacros(preprocess(read('preamble.js'), 'preamble.js'));
+    const preFile = MINIMAL_RUNTIME ? 'preamble_minimal.js' : 'preamble.js';
+    includeFile(preFile);
+
+    for (const item of libraryItems.concat(postSets)) {
+      print(indentify(item || '', 2));
     }
-    print(pre);
 
-    // Print out global variables and postsets TODO: batching
-    runJSify(true);
-
-    const generated = itemsDict.functionStub.concat(itemsDict.globalVariablePostSet);
-    generated.forEach((item) => print(indentify(item.JS || '', 2)));
-
-    if (USE_PTHREADS) {
+    if (PTHREADS) {
       print('\n // proxiedFunctionTable specifies the list of functions that can be called either synchronously or asynchronously from other threads in postMessage()d or internally queued events. This way a pthread in a Worker can synchronously access e.g. the DOM on the main thread.');
       print('\nvar proxiedFunctionTable = [' + proxiedFunctionTable.join() + '];\n');
     }
 
-    if (!MINIMAL_RUNTIME) {
-      print('var ASSERTIONS = ' + !!ASSERTIONS + ';\n');
-    }
-
     if ((SUPPORT_BASE64_EMBEDDING || FORCE_FILESYSTEM) && !MINIMAL_RUNTIME) {
-      print(preprocess(read('base64Utils.js')));
+      includeFile('base64Utils.js');
     }
 
-    if (abortExecution) throw Error('Aborting compilation due to previous errors');
+    if (abortExecution) {
+      throw Error('Aborting compilation due to previous errors');
+    }
 
     // This is the main 'post' pass. Print out the generated code that we have here, together with the
     // rest of the output that we started to print out earlier (see comment on the
@@ -521,41 +570,51 @@ function ${name}(${args}) {
 
     if (HEADLESS) {
       print('if (!ENVIRONMENT_IS_WEB) {');
-      print(read('headlessCanvas.js'));
-      print('\n');
-      print(read('headless.js').replace("'%s'", "'http://emscripten.org'").replace("'?%s'", "''").replace("'?%s'", "'/'").replace('%s,', 'null,').replace('%d', '0'));
+      includeFile('headlessCanvas.js');
+      includeFile('headless.js')
       print('}');
     }
     if (PROXY_TO_WORKER) {
       print('if (ENVIRONMENT_IS_WORKER) {\n');
-      print(read('webGLWorker.js'));
-      print(processMacros(preprocess(read('proxyWorker.js'), 'proxyWorker.js')));
+      includeFile('webGLWorker.js');
+      includeFile('proxyWorker.js');
       print('}');
     }
     if (DETERMINISTIC) {
-      print(read('deterministic.js'));
+      includeFile('deterministic.js');
     }
 
     const postFile = MINIMAL_RUNTIME ? 'postamble_minimal.js' : 'postamble.js';
-    const post = processMacros(preprocess(read(postFile), postFile));
-    print(post);
+    includeFile(postFile);
 
-    print(processMacros(preprocess(shellParts[1], shellFile)));
+    for (const fileName of POST_JS_FILES) {
+      includeFileRaw(fileName);
+    }
 
-    print('\n//FORWARDED_DATA:' + JSON.stringify({
+    print('//FORWARDED_DATA:' + JSON.stringify({
       librarySymbols: librarySymbols,
       warnings: warnings,
+      asyncFuncs,
       ATINITS: ATINITS.join('\n'),
-      ATMAINS: ATMAINS.join('\n'),
+      ATMAINS: STRICT ? '' : ATMAINS.join('\n'),
       ATEXITS: ATEXITS.join('\n'),
     }));
   }
 
-  // Data
-
-  if (mainPass) {
-    functionStubs.forEach(functionStubHandler);
+  for (const sym of symbolsNeeded) {
+    symbolHandler(sym);
   }
 
-  finalCombiner();
+  if (symbolsOnly) {
+    print(JSON.stringify({
+      deps: symbolDeps,
+      asyncFuncs
+    }));
+  } else {
+    finalCombiner();
+  }
+
+  if (abortExecution) {
+    throw Error('Aborting compilation due to previous errors');
+  }
 }
