@@ -5,17 +5,32 @@
  */
 
 #if USE_CLOSURE_COMPILER
+
 // if (!Module)` is crucial for Closure Compiler here as it will
 // otherwise replace every `Module` occurrence with the object below
 var /** @type{Object} */ Module;
-if (!Module) /** @suppress{checkTypes}*/Module = {"__EMSCRIPTEN_PRIVATE_MODULE_EXPORT_NAME_SUBSTITUTION__":1};
-#elif ENVIRONMENT_MAY_BE_NODE || ENVIRONMENT_MAY_BE_SHELL
+if (!Module) /** @suppress{checkTypes}*/Module = 
+#if AUDIO_WORKLET
+  globalThis.{{{ EXPORT_NAME }}} || 
+#endif
+  {"__EMSCRIPTEN_PRIVATE_MODULE_EXPORT_NAME_SUBSTITUTION__":1};
+
+#elif !MODULARIZE && (ENVIRONMENT_MAY_BE_NODE || ENVIRONMENT_MAY_BE_SHELL)
+
 // When running on the web we expect Module to be defined externally, in the
 // HTML.  Otherwise we must define it here before its first use
-var Module = typeof {{{ EXPORT_NAME }}} != 'undefined' ? {{{ EXPORT_NAME }}} : {};
+var Module =
+#if SUPPORTS_GLOBALTHIS
+  // As a small code size optimization, we can use 'globalThis' to refer to the global scope Module variable.
+  globalThis.{{{ EXPORT_NAME }}} || {};
+#else
+  // Otherwise do a good old typeof check.
+  typeof {{{ EXPORT_NAME }}} != 'undefined' ? {{{ EXPORT_NAME }}} : {};
+#endif
+
 #else
 var Module = {{{ EXPORT_NAME }}};
-#endif // USE_CLOSURE_COMPILER
+#endif
 
 #if MODULARIZE && EXPORT_READY_PROMISE
 // Set up the promise that indicates the Module is initialized
@@ -37,7 +52,11 @@ var ENVIRONMENT_IS_NODE = typeof process == 'object';
 var ENVIRONMENT_IS_SHELL = typeof read == 'function';
 #endif
 
-#if ASSERTIONS || USE_PTHREADS
+#if AUDIO_WORKLET
+var ENVIRONMENT_IS_AUDIO_WORKLET = typeof AudioWorkletGlobalScope !== 'undefined';
+#endif
+
+#if ASSERTIONS || PTHREADS
 #if !ENVIRONMENT_MAY_BE_NODE && !ENVIRONMENT_MAY_BE_SHELL
 var ENVIRONMENT_IS_WEB = true
 #elif ENVIRONMENT && !ENVIRONMENT.includes(',')
@@ -49,7 +68,7 @@ var ENVIRONMENT_IS_WEB = !ENVIRONMENT_IS_SHELL;
 #else
 var ENVIRONMENT_IS_WEB = !ENVIRONMENT_IS_NODE;
 #endif
-#endif // ASSERTIONS || USE_PTHREADS
+#endif // ASSERTIONS || PTHREADS
 
 #if WASM_WORKERS
 var ENVIRONMENT_IS_WASM_WORKER = Module['$ww'];
@@ -75,7 +94,7 @@ if (ENVIRONMENT_IS_NODE) {
   Module['wasm'] = fs.readFileSync(__dirname + '/{{{ TARGET_BASENAME }}}.wasm');
 #endif
 #endif
-#if MEM_INIT_METHOD == 1 && !MEM_INIT_IN_WASM
+#if !MEM_INIT_IN_WASM
   Module['mem'] = fs.readFileSync(__dirname + '/{{{ TARGET_BASENAME }}}.mem');
 #endif
 }
@@ -91,7 +110,7 @@ if (ENVIRONMENT_IS_SHELL) {
   Module['wasm'] = read('{{{ TARGET_BASENAME }}}.wasm', 'binary');
 #endif
 #endif
-#if MEM_INIT_METHOD == 1 && !MEM_INIT_IN_WASM
+#if !MEM_INIT_IN_WASM
   Module['mem'] = read('{{{ TARGET_BASENAME }}}.mem', 'binary');
 #endif
 }
@@ -119,6 +138,12 @@ function ready() {
 #elif ASSERTIONS
   out('ready() called, and INVOKE_RUN=0. The runtime is now ready for you to call run() to invoke application _main(). You can also override ready() in a --pre-js file to get this signal as a callback')
 #endif
+#if PTHREADS
+  // This Worker is now ready to host pthreads, tell the main thread we can proceed.
+  if (ENVIRONMENT_IS_PTHREAD) {
+    startWorker(Module);
+  }
+#endif
 }
 
 #if POLYFILL
@@ -130,10 +155,9 @@ function ready() {
 
 // --pre-jses are emitted after the Module integration code, so that they can
 // refer to Module (if they choose; they can also define Module)
+{{{ preJS() }}}
 
-// {{PRE_JSES}}
-
-#if USE_PTHREADS
+#if PTHREADS
 
 #if !MODULARIZE
 // In MODULARIZE mode _scriptDir needs to be captured already at the very top of the page immediately when the page is parsed, so it is generated there
@@ -151,6 +175,4 @@ var ENVIRONMENT_IS_WORKER = ENVIRONMENT_IS_PTHREAD = typeof importScripts == 'fu
 #endif
 
 var currentScriptUrl = typeof _scriptDir != 'undefined' ? _scriptDir : ((typeof document != 'undefined' && document.currentScript) ? document.currentScript.src : undefined);
-#endif // USE_PTHREADS
-
-{{BODY}}
+#endif // PTHREADS
